/*
* Change only colors, do not add more class, order or remove 
* also dont chage selectors definition
*/

.base {
    fill: #000000;
}

.success {
    fill: #009909;
}

.warning {
    fill: #FF1990;
}

.error {
    fill: #0099E6;
}

.dark .base {
    fill: #EBEBEB;
}

.dark .success {
    fill: #00990A;
}

.dark .warning {
    fill: #FF1990;
}

.dark .error {
    fill: #0099E6;
}